#!/usr/bin/python
# -*- coding: utf-8 -*-
'''
 *  Copyright (C) 2007 Karlheinz Klingbeil (lunqual)
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MeRCHANTABILITY or FITNeSS FOR A PARTICULAR PURPOSe.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 '''
 
 #------------------------------------------------
 # 
 # Hilfsklasse zur Berechnung von
 # Kontraktionsdaten
 # das Modul "tafelrechner.py" muss im
 # PYTHONPATH oder im gleichen Verzeichnis
 # vorhanden sein !
 #
 #------------------------------------------------
from tafelrechner import Tafel
#------------------------------------------
#
# Konstanten
#
#------------------------------------------

VERSION='Z0.4/01.10.2007'

 # die folgenden Namen  werden
 # als Schlüssel für eine einzelne Zutat
 # benötigt
NAME = 'Name'
STAERKE = 'Staerke'
LITER ='Liter'
TROCKENSUBSTANZ = 'Trockensubstanz'
LITERGEWICHT = 'Litergewicht'
LITER_EXTRAKTFREI = 'LiterExtraktfrei'
STAERKE_EXTRAKTFREI='StaerkeExtraktfrei'
LITERGEWICHT_EXTRAKTFREI = 'LitergewichtExtraktFrei'
LA='La'
GEWICHT_EXTRAKTFREI='GewichtExtraktfrei'
GEWICHT='Gewicht'
#--------------------------------------
# die folgenden Namen sind die
# Schlüssel für die gesamte
# Zutatenliste
#--------------------------------------

SUMME_LITER = 'SummeLiter'
SUMME_LITER_EXTRAKTFREI = 'SummeLiterExtraktfrei'
SUMME_GEWICHT_EXTRAKTFREI = 'SummeGewichtExtraktfrei'
SUMME_LA='SummeLA'
SUMME_STAERKE_EXTRAKTFREI='StaerkeExtraktfrei'
SUMME_GEWICHT_EXTRAKTFREI='GewichtExtraktfrei'
SUMME_GEWICHT_LITER='GewichtLiter'
KONTRAKTION_KG='KontraktionKg'
KONTRAKTION_LITER='KontraktionLiter'
ENDMENGE_LITER='EndmengeLiter'
ENDMENGE_STAERKE='EndmengeStaerke'
ENDMENGE_GEWICHT='EndmengeGewicht'

#------------------------------------------
#
# Zutat, die zum herabsetzen verwendet wird
#
#------------------------------------------
NAME_ZUTAT='-Verschnittwasser-'
EINSTELLEN_GENAUIGKEIT=0.00001
MAX_ITERATION = 100
#------------------------------------------
# Hilfsfunktion zur Prüfung auf Null
# nötig, weil wir mit Floating Point
# Arithmetik hantieren
#------------------------------------------

GENAUIGKEIT = 0.0001
def isNull(value):
    if abs(value) < GENAUIGKEIT:
        return True
    else:
        return False

#-------------------------------------------
#
# Hilfsfunktion zur Berechnung der
# LA aus Stärke und Liter
#
#------------------------------------------
def LA(Staerke=0.0,Liter=0.0):
    return Liter * (Staerke/100)


#------------------------------------------
#
# Eine einzelne Zutat
# Wir geben an:
#     - Name der Zutat
#     - die Anzahl der Liter
#     - die Alkoholstärke
#     - Trockensubstanz
#     - und das Litergewicht
#
# wobei das Litergewicht 
# berechnet wird, wenn die
# Trockensubstanz == Null ist.
#-----------------------------------------
class Zutat:
    daten = {}
    tafel = Tafel()
    
    #---------------------------------------
    # wir müssen auch auf die Daten
    # zugreifen können
    #--------------------------------------
    def get(self,key):
        return self.daten[key]
    
    def set(self,key,value):
        self.daten[key]=value
    
    def __init__(self,Name='Zutat',Staerke=0.0,Liter=0.0,Trockensubstanz=0.0,Litergewicht=tafel.Litergewicht(0.0,20)):
        self.daten = {}
        self.daten[NAME] = Name
        self.daten[STAERKE] = Staerke
        self.daten[LITER] = Liter
        self.daten[TROCKENSUBSTANZ] = Trockensubstanz
        #--------------------------------------
        # Wenn keine Trockensubstanz 
        # angegeben ist, wird das
        # extraktfreie Litergewicht
        # berechnet, ansonsten wird
        # das angegebene Litergewicht
        # verwendet
        #--------------------------------------
        if isNull(Trockensubstanz):
            self.daten[LITERGEWICHT] = self.tafel.Litergewicht(Staerke,20)
        else:
            self.daten[LITERGEWICHT] = Litergewicht
        self.rechnen()
    #---------------------------------------
    # Eine einzelne Zutat berechnen
    # Im einzelen müssen wir die
    # Liter extraktfreie Flüssigkeit,
    # deren Alkoholstärke und
    # Litergewicht ermitteln
    # Ausserdem ermitteln wir noch die
    # LA (Liter reiner Alkohol) der Zutat
    #---------------------------------------
    def rechnen(self):
        LiterExtraktfrei = 0.0
        StaerkeExtraktfrei = 0.0
        LitergewichtExtraktfrei = 0.0
        Gewicht = 0.0
        # später benötigen wir die LA der Zutat
        La = LA(Staerke=self.daten[STAERKE],Liter=self.daten[LITER])
        Gewicht = self.daten[LITER] * self.daten[LITERGEWICHT] 
        #-------------------------------------------------
        # Zuerst prüfen wir, ob wir es mit
        # einer extraktfreien Zutat zu tun 
        # haben, das vereinfacht die Sache
        #--------------------------------------------------
        if isNull(self.daten[TROCKENSUBSTANZ]):
            # Zutat ist extraktfrei
            # also ist die extraktfreie Flüssigkeitsmenge
            # gleich die Literzahl
            # die Stärke und das Litergewicht werden
            # einfach übernommen
            LiterExtraktfrei  = self.daten[LITER]
            StaerkeExtraktfrei = self.daten[STAERKE]
            LitergewichtExtraktfrei = self.tafel.Litergewicht(self.daten[STAERKE],20)
        else: 
            #-------------------------------------------------------
            # Zutat ist extrakthaltig
            # zuerst rechnen wir mit Hilfe 
            # der Trockensubstanzangabe
            # Menge des Extrakts in Kg heraus
            # und ermitteln so erst mal die
            # Menge der Flüssigkeit in KG
            #-------------------------------------------------------
            # wir haben das Gesamtgewicht der Zutat bereits oben
            # berechnet,
            # jetzt ermitteln wir das gewicht der Trockenmasse
            trockenmasse = Gewicht * (self.daten[TROCKENSUBSTANZ] / 100.0)
            # und zuletzt das Gewicht der Flüssigkeit
            gewichtFluessigkeit = Gewicht - trockenmasse
            #------------------------------------------------
            # als nächstes prüfen wir, ob die Zutat
            # Alkoholfrei ist oder nicht
            #------------------------------------------------
            if isNull(self.daten[STAERKE] ):
                # Zutat ist alkoholfrei
                # in diesem Fall besteht die Flüssigkeit
                # aus Wasser, mit
                # bekanntem Litergewicht
                StaerkeExtraktfrei = 0.0
                LitergewichtExtraktfrei = self.tafel.Litergewicht(0,20.0)
                LiterExtraktfrei = gewichtFluessigkeit * (1/LitergewichtExtraktfrei)
            else:
                # Zutat hat Extrakt UND Alkohol
                # die Flüssigkeit besteht deshalb
                # aus einer Alkohol-Wasser Mischung.
                #
                # Da das Litergewicht von reinem Alkohol
                # bekannt ist, ermitteln wir zunächst das 
                # Gewicht des enthaltenen Alkohols 
                # (Die LA haben wir schon vorher berechnet)
                GewichtAlkohol = La * self.tafel.Litergewicht(100,20)
                # da wir nun das Gewicht des Alkohols und
                # das Gesamtgewicht der Flüssigkeit kennen,
                # haben wir die Massenprozent
                masProz = (GewichtAlkohol * 100) / gewichtFluessigkeit
                # Damit können wir die Alkoholstärke in
                # %Vol nach Tafel5 ermitteln
                StaerkeExtraktfrei = self.tafel.Tafel5(masProz)
                # und mit der Stärke in %Vol das Litergewicht
                # der extraktfreien Flüssigkeit, und damit
                # die Liter extraktfreier Flüssigkeit
                LitergewichtExtraktfrei = self.tafel.Litergewicht(StaerkeExtraktfrei,20)
                LiterExtraktfrei = gewichtFluessigkeit * (1/LitergewichtExtraktfrei)


        #-------------------------------
        # Die ermittelten Daten ins Dictionary eintragen
        #-------------------------------
        self.daten[LITER_EXTRAKTFREI] = LiterExtraktfrei
        self.daten[STAERKE_EXTRAKTFREI] = StaerkeExtraktfrei
        self.daten[LITERGEWICHT_EXTRAKTFREI] = LitergewichtExtraktfrei
        self.daten[LA] = La
        self.daten[GEWICHT_EXTRAKTFREI] = LiterExtraktfrei * LitergewichtExtraktfrei
        self.daten[GEWICHT] = Gewicht
    #----------------------------------------
    # Daten als String ausgeben
    #---------------------------------------- 
    def toString(self):
        r = ''
        r += '\n%s\n' %('+' * 70)
        r +=  'Name=%s, Staerke=%2.2f %%Vol, Liter=%12.4f Liter\n' % \
               (self.daten[NAME],self.daten[STAERKE],self.daten[LITER])
        r+= 'Trockensubstanz=%12.4f %%mas, Litergewicht=%2.6f kg/Liter\n' % \
                (self.daten[TROCKENSUBSTANZ],self.daten[LITERGEWICHT])
        r += 'LA = %12.4f LA, LiterExtraktfrei=%12.4f Liter\n' % \
                (self.daten[LA],self.daten[LITER_EXTRAKTFREI])
        r += 'StärkeExtraktfrei=%2.2f %%Vol ,\nLitergewichtExtraktfrei=%2.6f kg/Liter\n' % \
                (self.daten[STAERKE_EXTRAKTFREI],self.daten[LITERGEWICHT_EXTRAKTFREI])
        r += 'Gewicht Zutat=%12.4f Kg, Gewicht extraktfrei=%12.4f Kg\n' % \
                (self.daten[GEWICHT],self.daten[GEWICHT_EXTRAKTFREI])
        return r
    #-----------------------------------------
    # Version zurückgeben
    #-----------------------------------------
    def version(self):
        return VERSION  
        
#------------------------------------------------
#
# Zutatenliste
#
#------------------------------------------------
class ZutatenListe:
    liste = {}
    summe={}
    tafel = Tafel()
    def __init__(self):
        self.liste = {}
        self.summe = {}
        
    def getListe(self,key):
        return self.liste[key]

    def getSumme(self,key):
        return self.summe[key]
    #-------------------------------
    # eine Zutat hinzufügen
    #-------------------------------
    def add(self,zutat):
        self.liste[zutat.get(NAME)] = zutat
        
    #-----------------------------
    # eine Zutat entfernen
    #-----------------------------
    def delete(self,name):
        del self.liste[name]
        
    #----------------------------
    # die Kontraktion ausrechnen
    # das funktioniert so, dass wir
    # die Daten für jede Zutat errechnen und
    # aufsummieren
    #----------------------------
    def rechnen(self):
        summeLiter = 0.0
        summeLa= 0.0
        summeLiterExtraktfrei = 0.0
        summeGewichtExtraktfrei = 0.0
        summeGewichtLiter = 0.0
        summeGewicht = 0.0
        staerkeExtraktfrei = 0.0
        kontraktionKg=0.0
        kontraktionLiter=0.0
        endmengeLiter=0.0
        endmengeGewicht = 0.0
        endmengeStaerke=0.0
        #--------------------------------
        # Wir gehen nun alle Zutaten durch 
        # und summieren auf
        #---------------------------------
        for zutat in self.liste.items():
            zutat = zutat[1]
            # die Zutat ausrechnen
            zutat.rechnen()
            summeLiter += zutat.get(LITER)
            summeLa += zutat.get(LA)
            summeLiterExtraktfrei += zutat.get(LITER_EXTRAKTFREI)
            summeGewichtExtraktfrei += zutat.get(GEWICHT_EXTRAKTFREI)
            summeGewicht += zutat.get(GEWICHT)
        #------------------------------
        # Die errechneten Werte übernehmen
        # Vorsicht ist die Mutter der Porzellankiste !
        #-----------------------------
        try:
            staerkeExtraktfrei= (summeLa/summeLiterExtraktfrei) * 100
            summeGewichtLiter = summeLiterExtraktfrei * self.tafel.Litergewicht(staerkeExtraktfrei,20.0)
            kontraktionKg = abs(summeGewichtExtraktfrei - summeGewichtLiter)
            kontraktionLiter = kontraktionKg * (1/self.tafel.Litergewicht(0.0,20.0))
            endmengeLiter = summeLiter - kontraktionLiter
            endmengeStaerke = (summeLa * 100) / endmengeLiter
            endmengeGewicht = summeGewicht
        except:
            print('Fehler bei der Berechnung !\nTeilung durch Null oder keine Zutaten angegeben !\n')              
        self.summe[SUMME_LITER] = summeLiter
        self.summe[SUMME_LA] = summeLa
        self.summe[SUMME_LITER_EXTRAKTFREI] = summeLiterExtraktfrei
        self.summe[SUMME_STAERKE_EXTRAKTFREI] = staerkeExtraktfrei
        self.summe[SUMME_GEWICHT_EXTRAKTFREI] = summeGewichtExtraktfrei
        self.summe[SUMME_GEWICHT_LITER] = summeGewichtLiter 
        self.summe[KONTRAKTION_KG] = kontraktionKg
        self.summe[KONTRAKTION_LITER] =  kontraktionLiter
        self.summe[ENDMENGE_LITER] =endmengeLiter
        self.summe[ENDMENGE_STAERKE] = endmengeStaerke
        self.summe[ENDMENGE_GEWICHT] = endmengeGewicht
    
    def einstellen(self,ziel):
        # erstmal ausrechnen, was wir an Stärke haben
        self.rechnen()
        # und jetzt ausrechnen, was wir an Wasser Theoretisch 
        # benötigen
        staerke = self.summe[ENDMENGE_STAERKE]
        # Mischungskreuz bildern
        # Differenz1
        d1 = abs(ziel)
        d2 = abs(staerke - ziel)
        verhaeltnis=d2/d1
        liter = Liter=self.summe[ENDMENGE_LITER] * verhaeltnis
        liter_alt=liter
        # nun müssen wir noch eine zusätzliche Zutat unterbringen
        zutat = Zutat(Name=NAME_ZUTAT,Liter=liter)
        self.add(zutat)
        self.rechnen()
        unten = liter_alt
        oben = liter_alt + self.summe[KONTRAKTION_LITER]  * 2
        mitte = (unten + oben) /2
        ende=False
        zutat.set(LITER,mitte)
        self.rechnen()
        i=0
        #--------------------------------
        #
        # Da jedes Hinzugeben von Wasser seinerseits
        # die Kontraktion erhöht und damit wiederum
        # weitere Kontraktion auslösst, müssen wir
        # uns der endgültigen benötigten Wassermenge
        # iterativ annähern
        #
        #--------------------------------
        while not ende:
            i += 1
            if abs(self.summe[ENDMENGE_STAERKE] - ziel) < EINSTELLEN_GENAUIGKEIT or i > MAX_ITERATION:
                ende =True
            else:
                if self.summe[ENDMENGE_STAERKE] > ziel:
                    unten = zutat.get(LITER)
                else:
                    oben = zutat.get(LITER)
                mitte = (unten + oben)/2
                zutat.set(LITER, mitte)
                if  isNull(oben -unten):
                    ende = True
                self.rechnen()                
                        
    def toString(self):
        r= ''
        for zutat in self.liste.items():
            r += '\n%s' %(zutat[1].toString())
            
        r += '\n%s\n' %('_' * 70)
        r += 'Summe Liter = %12.4f Liter\n' %(self.summe[SUMME_LITER])
        r += 'Summe extraktfreie Flüssigkeit = %12.4f Liter\n' %(self.summe[SUMME_LITER_EXTRAKTFREI])
        r += 'Alkoholstärke extraktfrei = %2.4f %%Vol\n' %(self.summe[SUMME_STAERKE_EXTRAKTFREI])
        r += 'Gewicht extraktfrei = %12.4f kg\n' %(self.summe[SUMME_GEWICHT_EXTRAKTFREI])
        r += 'Gewicht über Liter = %12.4f kg\n' %(self.summe[SUMME_GEWICHT_LITER])
        r += '\n\nKontraktion = %12.4f kg\n' %(self.summe[KONTRAKTION_KG])
        r += 'Kontraktion = %12.4f Liter\n' %(self.summe[KONTRAKTION_LITER])
        r += '\n\nEndmenge = %12.4f Liter\n' %(self.summe[ENDMENGE_LITER])
        r += 'Endmenge = %12.4f Kg\n' %(self.summe[ENDMENGE_GEWICHT])
        r += 'Alkoholstärke = %2.4f %%Vol\n' %(self.summe[ENDMENGE_STAERKE])
   
           
        r += '\n%s\n' %('_' * 70)
        
        return r
    
    
